# gradient_coding
# This is an implementation of https://arxiv.org/abs/1612.03301 for logistic regression by the original author

# For usage instructions, please see the associated wiki (https://github.com/rashisht1/gradient_coding/wiki).

from __future__ import print_function

import os
import random
from re import X
import sys
import time
import argparse

import numpy as np
import scipy.sparse as sps

# import scipy.special as sp
from sklearn.preprocessing import OneHotEncoder
from mpi4py import MPI
import matplotlib.pyplot as plt
import pandas as pd
from tqdm.auto import tqdm
from scipy.special import softmax

from util import *

ohe = OneHotEncoder(sparse=False)


def coded_logistic_regression(
    n_procs,
    n_samples,
    n_features,
    input_dir,
    n_stragglers,
    is_real_data,
    params,
    n_classes,
    n_hidden=24,
):

    comm = MPI.COMM_WORLD
    rank = comm.Get_rank()
    size = comm.Get_size()

    rounds = params[0]
    layer_0_dim = (n_features, n_hidden)
    layer_1_dim = (n_hidden, n_classes)
    flat_dim = (
        layer_0_dim[0] * layer_0_dim[1] + layer_1_dim[0] * layer_1_dim[1]
    )
    n_workers = n_procs - 1
    rows_per_worker = n_samples // (n_procs - 1)
    np.random.seed(42)
    # beta_message = np.zeros(flat_dim)

    # Loading the data
    if rank:

        y = ohe.fit_transform(
            load_data(input_dir + "label.dat").reshape((-1, 1))
        )
        X_current = np.zeros(
            [(1 + n_stragglers) * rows_per_worker, n_features]
        )
        y_current = np.zeros(((1 + n_stragglers) * rows_per_worker, n_classes))
        for i in range(1 + n_stragglers):
            X_current[
                i * rows_per_worker : (i + 1) * rows_per_worker, :
            ] = load_data(
                input_dir + str((rank - 1 + i) % n_workers + 1) + ".dat"
            )
            y_current[i * rows_per_worker : (i + 1) * rows_per_worker] = y[
                ((rank - 1 + i) % n_workers)
                * rows_per_worker : ((rank - 1 + i) % n_workers + 1)
                * rows_per_worker
            ]
    # Initializing relevant variables
    B = np.zeros((n_workers, n_workers))
    beta0 = np.random.randn(n_features, n_hidden) / np.sqrt(n_features)
    beta1 = np.random.randn(n_hidden, n_classes) / np.sqrt(n_hidden)
    beta_message_from_master = message_flat(
        beta0, beta1, layer_0_dim, layer_1_dim
    )
    # if rank:
    g_logits_w0_helper = np.zeros(
        (n_samples, args.features, n_hidden, n_classes)
    )
    g_loss_w0_helper = np.zeros(
        (n_samples, args.features, n_hidden, n_classes)
    )

    if rank:
        in_layer_1 = X_current.dot(beta0)

        # Z_i = relu(A_i), dimensions = (n_rows(X_current), hidden_nodes)
        out_layer_1 = relu(in_layer_1)

        # B_i = Z_i . W_i^1, dimensions = (n_rows(X_current), outputs)
        logits = out_layer_1.dot(beta1)

        # dimensions = (n_rows(X_current), outputs)
        y_hat = softmax(logits, axis=1)

        # shared by everyone
        # dimesnions = (n_rows(X_current), outputs)
        # g_loss_yhat = np.divide(
        #     -y_current * np.exp(-np.multiply(y_hat, y_current)),
        #     1 + np.exp(-np.multiply(y_hat, y_current)),
        # )

        #
        # dimensions = (n_rows(X_current), outputs)
        # g_yhat_logits = sigmoid_derivative(logits)
        g_loss_logits = y_hat - y_current
        # w1 computations

        # dimensions = (n_rows(X_current), hidden_nodes)
        g_logits_w1 = out_layer_1

        # dimensions = (hidden_nodes, outputs)
        # we did a trick to sum along datapoints to get correct dimensions
        # TODO need to do sum B.row (see original code)
        g_loss_w1 = g_logits_w1.T.dot(
            g_loss_logits
        )  # g_loss_yhat * g_yhat_logits)

        # w0 computations
        # g_logits_w0 = np.zeros((args.features, hidden_nodes, outputs))

        # for k in range(X_current.shape[0]):
        #     for j in range(n_hidden):
        #         g_logits_w0_helper[k, :, j] = np.outer(
        #             X_current[k],
        #             (relu_derivative(in_layer_1[k, j]) * (beta1[j])),
        #         )
        #         g_loss_w0_helper[k, :, j] = (
        #             g_logits_w0_helper[k, :, j]
        #             * g_loss_yhat[k]
        #             * g_yhat_logits[k]
        #         )

        g_loss_w0 = np.einsum(
            "ki,kj,jl,kl->ij",
            X_current,
            relu_derivative(in_layer_1),
            beta1,
            g_loss_logits
            # g_loss_yhat,
            # g_yhat_logits,
        )

        # TODO need to do sum B.row (see original code)
        # g_loss_w0 = np.einsum("ijkl->jk", g_loss_w0_helper)

        g0 = g_loss_w0
        g1 = g_loss_w1

        # -X_current.T.dot(
        # np.divide(y_current, np.exp(np.multiply(predy, y_current)) + 1)
        # )
        send_req = MPI.Request()
        recv_reqs = []

    else:

        B = getB(n_workers, n_stragglers)
        # A = np.zeros((int(sp.binom(n_workers,n_stragglers)),n_workers))
        # A=getA(B,n_workers,n_stragglers)

        msgBuffers = np.array([np.zeros(flat_dim) for i in range(n_procs - 1)])
        g0_helper = np.zeros((msgBuffers.shape[0], n_features, n_hidden))
        g1_helper = np.zeros((msgBuffers.shape[0], n_hidden, n_classes))
        g0 = np.zeros((n_features, n_hidden))
        g1 = np.zeros((n_hidden, n_classes))

        A_row = np.zeros((1, n_procs - 1))

        beta0set = np.zeros((rounds, n_features, n_hidden))
        beta1set = np.zeros((rounds, n_hidden, n_classes))
        timeset = np.zeros(rounds)
        worker_timeset = np.zeros((rounds, n_procs - 1))

        request_set = []
        recv_reqs = []
        send_set = []

        cnt_completed = 0
        completed_workers = np.ndarray(n_procs - 1, dtype=bool)
        status = MPI.Status()

        eta0 = params[2]  # ----- learning rate
        alpha = params[1]  # --- coefficient of l2 regularization
        utemp = np.zeros(n_features)  # for accelerated gradient descent

    B = comm.bcast(B, root=0)

    # Setting up y_current_mod on all workers
    if rank:
        y_current_mod = np.zeros(
            ((1 + n_stragglers) * rows_per_worker, n_classes)
        )
        for i in range(1 + n_stragglers):
            y_current_mod[i * rows_per_worker : (i + 1) * rows_per_worker] = (
                B[rank - 1, ((rank - 1 + i) % n_workers)]
                * y_current[i * rows_per_worker : (i + 1) * rows_per_worker]
            )

    # Posting all Irecv requests for master and workers
    if rank:

        for i in range(rounds):
            # worker recv
            req = comm.Irecv(
                [beta_message_from_master, MPI.DOUBLE], source=0, tag=i
            )
            beta0, beta1 = message_unflat(
                beta_message_from_master, layer_0_dim, layer_1_dim
            )

            recv_reqs.append(req)
    else:

        for i in range(rounds):
            recv_reqs = []
            for proc in range(1, n_procs):
                # master recv
                req = comm.Irecv(
                    [msgBuffers[proc - 1], MPI.DOUBLE], source=proc, tag=i
                )
                recv_reqs.append(req)
            request_set.append(recv_reqs)

    #######################################################################################################
    comm.Barrier()
    if rank == 0:
        print(
            "---- Starting Coded Iterations for "
            + str(n_stragglers)
            + " stragglers ----"
        )
        orig_start_time = time.time()

    for i in range(rounds):

        if rank == 0:

            if i % 10 == 0:
                print("\t >>> At Iteration %d" % (i))

            A_row[:] = 0
            send_set[:] = []
            completed_workers[:] = False
            cnt_completed = 0

            start_time = time.time()

            beta_message = message_flat(beta0, beta1, layer_0_dim, layer_1_dim)

            for l in range(1, n_procs):
                # master send
                sreq = comm.Isend([beta_message, MPI.DOUBLE], dest=l, tag=i)
                send_set.append(sreq)

            while cnt_completed < n_procs - 1 - n_stragglers:
                req_done = MPI.Request.Waitany(request_set[i], status)
                src = status.Get_source()
                worker_timeset[i, src - 1] = time.time() - start_time
                request_set[i].pop(req_done)
                cnt_completed += 1
                completed_workers[src - 1] = True

            completed_ind_set = [
                l for l in range(n_procs - 1) if completed_workers[l]
            ]
            A_row[0, completed_ind_set] = np.linalg.lstsq(
                B[completed_ind_set, :].T, np.ones(n_workers)
            )[0]

            A_row_new = np.einsum("ij->j", A_row)

            for idx in range(msgBuffers.shape[0]):
                g0_helper[idx], g1_helper[idx] = message_unflat(
                    msgBuffers[idx], layer_0_dim, layer_1_dim
                )
            g0 = np.einsum("j,jkl->kl", A_row_new, g0_helper)  # np.squeeze()
            g1 = np.einsum("j,jkl->kl", A_row_new, g1_helper)  # np.squeeze()

            # case_idx = calculate_indexA(completed_stragglers)
            # g = np.dot(A[case_idx,ind_set],tmpBuff)

            grad_multiplier = eta0[i] / n_samples
            # ---- update step for gradient descent
            np.subtract(
                (1 - 2 * alpha * eta0[i]) * beta0,
                grad_multiplier * g0,
                out=beta0,
            )
            np.subtract(
                (1 - 2 * alpha * eta0[i]) * beta1,
                grad_multiplier * g1,
                out=beta1,
            )

            # ---- updates for accelerated gradient descent
            # theta = 2.0 / (i + 2.0)
            # ytemp = (1 - theta) * beta1 + theta * utemp
            # beta1temp = (
            #     ytemp - grad_multiplier * g - (2 * alpha * eta0[i]) * beta1
            # )
            # utemp = beta1 + (beta1temp - beta1) * (1 / theta)
            # beta1[:] = beta1temp

            timeset[i] = time.time() - orig_start_time
            beta0set[i, :] = beta0
            beta1set[i, :] = beta1

            ind_set = [
                l for l in range(n_procs - 1) if not completed_workers[l]
            ]
            for l in ind_set:
                worker_timeset[i, l] = -1

        else:
            recv_reqs[i].Wait()
            sendTestBuf = send_req.test()
            if not sendTestBuf[0]:
                send_req.Cancel()
                # print("Worker " + str(rank) + " cancelled send request for Iteration " + str(i))

            # predy = #X_current.dot(beta1)
            in_layer_1 = X_current.dot(beta0)

            # print(beta0, in_layer_1, X_current)

            # Z_i = relu(A_i), dimensions = (n_rows(X_current), hidden_nodes)
            out_layer_1 = relu(in_layer_1)

            # B_i = Z_i . W_i^1, dimensions = (n_rows(X_current), outputs)
            logits = out_layer_1.dot(beta1)

            # dimensions = (n_rows(X_current), outputs)
            y_hat = softmax(logits, axis=1)

            # shared by everyone
            # dimesnions = (n_rows(X_current), outputs)
            # g_loss_yhat = np.divide(
            #     -y_current * np.exp(-np.multiply(y_hat, y_current)),
            #     1 + np.exp(-np.multiply(y_hat, y_current)),
            # )

            # dimensions = (n_rows(X_current), outputs)
            # g_yhat_logits = sigmoid_derivative(logits)

            # w1 computations
            g_loss_logits = y_hat - y_current
            # dimensions = (n_rows(X_current), hidden_nodes)
            g_logits_w1 = out_layer_1

            # dimensions = (hidden_nodes, outputs)
            # we did a trick to sum along datapoints to get correct dimensions
            # TODO: do sum with B_row
            # g1_encoded = np.zeros(
            #     ((1 + n_stragglers) * rows_per_worker, n_hidden, n_classes)
            # )

            B_helper = np.zeros((X_current.shape[0],))

            for k in range(1 + n_stragglers):
                for qq in range(X_current.shape[0] // (1 + n_stragglers)):
                    B_helper[qq] = B[rank - 1, ((rank - 1 + k) % n_workers)]

            g_loss_w1 = np.einsum(
                "ij,ik,i->jk",
                g_logits_w1,
                g_loss_logits,
                # g_loss_yhat,
                # g_yhat_logits,
                B_helper,
            )

            # for k in range(X_current.shape[0]):
            #     g1_encoded[i * rows_per_worker : (i + 1) * rows_per_worker] = (
            #         B[rank - 1, ((rank - 1 + i) % n_workers)]
            #         * tmp[i * rows_per_worker : (i + 1) * rows_per_worker]
            #     )

            # g_loss_w1 = np.einsum("ijk->jk", g1_encoded)

            # w0 computations
            # g_logits_w0 = np.zeros((args.features, hidden_nodes, outputs))
            # g_logits_w0_helper = np.zeros(
            #     (X_current.shape[0], args.features, n_hidden, n_classes)
            # )
            # g_loss_w0_helper = np.zeros(
            #     (X_current.shape[0], args.features, n_hidden, n_classes)
            # )

            # g_logits_w0_helper = np.einsum(
            #     "ki,kj,jl->kijl",
            #     X_current,
            #     relu_derivative(in_layer_1),
            #     beta1,
            # )
            # g_loss_w0_helper = np.einsum(
            #     "kijl,kl,kl->kijl",
            #     g_logits_w0_helper,
            #     g_loss_yhat,
            #     g_yhat_logits,
            # )
            # for k in range(X_current.shape[0]):
            #     for j in range(n_hidden):
            #         g_logits_w0_helper[k, :, j] = np.outer(
            #             X_current[k],
            #             (relu_derivative(in_layer_1[k, j]) * (beta1[j])),
            #         )
            #         g_loss_w0_helper[k, :, j] = (
            #             g_logits_w0_helper[k, :, j]
            #             * g_loss_yhat[k]
            #             * g_yhat_logits[k]
            #         )

            # g_loss_w0_helper_helper = np.einsum("ijkl->ijk", g_loss_w0_helper)

            # # TODO: do sum with B_row
            # g0_encoded = np.zeros(
            #     ((1 + n_stragglers) * rows_per_worker, n_features, n_hidden)
            # )

            B_helper = np.zeros((X_current.shape[0],))

            for k in range(1 + n_stragglers):
                for qq in range(X_current.shape[0] // (1 + n_stragglers)):
                    B_helper[qq] = B[rank - 1, ((rank - 1 + k) % n_workers)]

            g_loss_w0 = np.einsum(
                "ki,kj,jl,kl,k->ij",
                X_current,
                relu_derivative(in_layer_1),
                beta1,
                g_loss_logits,
                # g_loss_yhat,
                # g_yhat_logits,
                B_helper,
            )
            # g0_encoded[k * rows_per_worker : (k + 1) * rows_per_worker] = (
            #     B[rank - 1, ((rank - 1 + k) % n_workers)]
            #     * g_loss_w0_helper_helper[
            #         k * rows_per_worker : (k + 1) * rows_per_worker
            #     ]
            # )

            # g_loss_w0 = np.einsum("ijk->jk", g0_encoded)
            g0 = g_loss_w0  # n_feauters x n_hidden, to encode must be: n_samples x n_hidden x n_classes
            g1 = g_loss_w1  # n_hidden x n_classes, to encode must be: n_samples x n_features x n_hidden

            # X_current.T.dot(
            #     np.divide(
            #         y_current_mod, np.exp(np.multiply(predy, y_current)) + 1
            #     )
            # )

            # g1 *= -1
            # g0 *= -1
            # worker send
            message = message_flat(g0, g1, layer_0_dim, layer_1_dim)
            send_req = comm.Isend([message, MPI.DOUBLE], dest=0, tag=i)

            # print(f"\n\n\n\n\n\n\nRank: {rank}, YO\n\n\n\n\n\n\n\n\n\n")

    #####################################################################################################
    comm.Barrier()
    if rank == 0:
        elapsed_time = time.time() - orig_start_time
        print("Total Time Elapsed: %.3f" % (elapsed_time))
        # Load all training data
        if not is_real_data:
            X_train = load_data(input_dir + "1.dat")
            print(">> Loaded 1")
            for j in range(2, n_procs - 1):
                X_temp = load_data(input_dir + str(j) + ".dat")
                X_train = np.vstack((X_train, X_temp))
                print(">> Loaded " + str(j))
        else:
            X_train = load_sparse_csr(input_dir + "1")
            for j in range(2, n_procs - 1):
                X_temp = load_sparse_csr(input_dir + str(j))
                X_train = sps.vstack((X_train, X_temp))

        y_train = ohe.fit_transform(
            load_data(input_dir + "label.dat").reshape((-1, 1))
        )
        y_train = y_train[0 : X_train.shape[0]]

        # Load all testing data
        y_test = ohe.fit_transform(
            load_data(input_dir + "label_test.dat").reshape((-1, 1))
        )
        if not is_real_data:
            X_test = load_data(input_dir + "test_data.dat")
        else:
            X_test = load_sparse_csr(input_dir + "test_data")

        n_train = X_train.shape[0]
        n_test = X_test.shape[0]

        training_loss = np.zeros(rounds)
        testing_loss = np.zeros(rounds)
        testing_acc = np.zeros(rounds)
        auc_loss = np.zeros(rounds)

        from sklearn.metrics import auc, roc_curve

        for i in range(rounds):
            beta1 = np.squeeze(beta1set[i, :])
            beta0 = np.squeeze(beta0set[i, :])
            predy_train = softmax(relu(X_train.dot(beta0)).dot(beta1), axis=1)
            predy_test = softmax(relu(X_test.dot(beta0)).dot(beta1), axis=1)
            training_loss[i] = np.mean(
                -y_train * np.log(predy_train)
            )  # calculate_loss(y_train, predy_train, n_train)
            testing_loss[i] = np.mean(
                -y_test * np.log(predy_test)
            )  # calculate_loss(y_test, predy_test, n_test)
            testing_acc[i] = (
                predy_test.argmax(axis=1) == y_test.argmax(axis=1)
            ).mean()
            # fpr, tpr, thresholds = roc_curve(y_test, predy_test, pos_label=1)
            # auc_loss[i] = auc(fpr, tpr)
            # print(
            #     "Iteration %d: Train Loss = %5.3f, Test Loss = %5.3f, AUC = %5.3f, Total time taken =%5.3f"
            #     % (
            #         i,
            #         training_loss[i],
            #         testing_loss[i],
            #         auc_loss[i],
            #         timeset[i],
            #     )
            # )

            print(
                "Iteration %d: Train Loss = %5.10f, Test Loss = %5.10f,  Test Acc = %5.10f, Total time taken =%5.3f"
                % (
                    i,
                    training_loss[i],
                    testing_loss[i],
                    testing_acc[i],
                    timeset[i],
                )
            )

        output_dir = input_dir + "results/"
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)

        save_vector(
            training_loss,
            output_dir + "coded_acc_%d_training_loss.dat" % (n_stragglers),
        )
        save_vector(
            testing_loss,
            output_dir + "coded_acc_%d_testing_loss.dat" % (n_stragglers),
        )
        output_csv = pd.DataFrame(columns=["testing_loss", "times"])

        output_csv["testing_loss"] = testing_loss
        output_csv["times"] = timeset - timeset[0]

        output_csv.to_csv("figures/gradient_coding.csv", index=False)
        plt.plot(testing_loss)
        plt.savefig("figures/grad_coding_testing_loss.png")
        plt.close()
        plt.plot(testing_acc)
        plt.savefig("figures/grad_coding_testing_acc.png")
        plt.close()
        # save_vector(
        #     auc_loss, output_dir + "coded_acc_%d_auc.dat" % (n_stragglers)
        # )
        save_vector(
            timeset, output_dir + "coded_acc_%d_timeset.dat" % (n_stragglers)
        )
        save_matrix(
            worker_timeset,
            output_dir + "coded_acc_%d_worker_timeset.dat" % (n_stragglers),
        )
        print(">>> Done")

    comm.Barrier()


parser = argparse.ArgumentParser()

parser.add_argument(
    "--samples", "-s", type=int, help=("Sample size"), default=100
)
parser.add_argument(
    "--features", "-f", type=int, help=("Number of features"), default=11
)
parser.add_argument(
    "--learning-rate",
    "-lr",
    type=float,
    help="Learning rate value. Default: 1e-2",
    default=1e-2,
)
parser.add_argument(
    "--regularizer",
    "-r",
    type=float,
    help="Regularizer value. Default: 1e-4",
    default=1e-4,
)
parser.add_argument(
    "--epochs",
    "-e",
    type=int,
    help="Number of epochs",
    default=10,
)
parser.add_argument("--num-nodes", "-nn", default=9, type=int)
parser.add_argument("--num-stragglers", "-ns", default=4, type=int)
parser.add_argument("--num-classes", "-nout", default=4, type=int)
parser.add_argument("--hidden-nodes", "-hn", default=10, type=int)

args = parser.parse_args()

coded_logistic_regression(
    args.num_nodes,
    args.samples,
    args.features,
    f"./artificial-data/{args.samples}x{args.features}/{args.num_nodes-1}/",
    args.num_stragglers,
    False,
    [
        args.epochs,
        args.regularizer,
        [args.learning_rate for _ in range(args.epochs)],
    ],
    args.num_classes,
    n_hidden=args.hidden_nodes,
)
