"""each worker trains on its own partition of data; 

workers only receive weights from master"""
import argparse

import numpy as np
from mpi4py import MPI
from sklearn.datasets import make_classification, make_regression

from nn.layers import Linear, ReLU
from nn.network import Network
from nn.loss import MSE
from nn.optimizer import SGD


parser = argparse.ArgumentParser()

parser.add_argument(
    "--type",
    "-t",
    type=str,
    help=(
        "Problem type, one of"
        "'r' for regression or"
        " 'c' for classification (binary)"
    ),
    default="c",
)
parser.add_argument(
    "--samples", "-s", type=int, help=("Sample size"), default=300
)
parser.add_argument(
    "--features", "-f", type=int, help=("Number of features"), default=9
)
parser.add_argument(
    "--learning-rate",
    "-lr",
    type=float,
    help="Learning rate value. Default: 1e-2",
    default=1e-2,
)
parser.add_argument(
    "--epochs",
    "-e",
    type=int,
    help="Number of epochs",
    default=10,
)
import numpy as np
from mpi4py import MPI
from sklearn.datasets import make_classification, make_regression

from nn.layers import Linear, ReLU
from nn.network import Network
from nn.loss import MSE
from nn.optimizer import SGD


parser = argparse.ArgumentParser()

parser.add_argument(
    "--type",
    "-t",
    type=str,
    help=(
        "Problem type, one of"
        "'r' for regression or"
        " 'c' for classification (binary)"
    ),
    default="c",
)
parser.add_argument(
    "--samples", "-s", type=int, help=("Sample size"), default=300
)
parser.add_argument(
    "--features", "-f", type=int, help=("Number of features"), default=9
)
parser.add_argument(
    "--learning-rate",
    "-lr",
    type=float,
    help="Learning rate value. Default: 1e-2",
    default=1e-2,
)
parser.add_argument(
    "--epochs",
    "-e",
    type=int,
    help="Number of epochs",
    default=10,
)
import numpy as np
from mpi4py import MPI
from sklearn.datasets import make_classification, make_regression

from nn.layers import Linear, ReLU
from nn.network import Network
from nn.loss import MSE
from nn.optimizer import SGD


parser = argparse.ArgumentParser()

parser.add_argument(
    "--type",
    "-t",
    type=str,
    help=(
        "Problem type, one of"
        "'r' for regression or"
        " 'c' for classification (binary)"
    ),
    default="c",
)
parser.add_argument(
    "--samples", "-s", type=int, help=("Sample size"), default=300
)
parser.add_argument(
    "--features", "-f", type=int, help=("Number of features"), default=9
)
parser.add_argument(
    "--learning-rate",
    "-lr",
    type=float,
    help="Learning rate value. Default: 1e-2",
    default=1e-2,
)
parser.add_argument(
    "--epochs",
    "-e",
    type=int,
    help="Number of epochs",
    default=10,
)


args = parser.parse_args()

if args.type == "r":
    X, y = make_regression(args.samples, args.features)
if args.type == "c":
    X, y = make_classification(args.samples, args.features)

net = Network(layers=[Linear(args.features, 1), ReLU()])
optimizer = SGD(lr=args.learning_rate)
loss_fn = MSE()
weights = np.random.randn(args.features, 1)

assert (
    X.shape[0] == y.shape[0]
), "Number of samples does not match between X and y"

comm = MPI.COMM_WORLD
name = MPI.Get_processor_name()
rank = comm.Get_rank()
TOTAL = comm.Get_size()

epoch = 1
if rank == 0:
    grads = dict()
    step = X.shape[0] // TOTAL

    for i in range(1, TOTAL):
        comm.send(X[(i - 1) * step : i * step], dest=i, tag=22)
        comm.send(y[(i - 1) * step : i * step], dest=i, tag=23)
        comm.send(net, dest=i, tag=33)

    while epoch < args.epochs:
        for i in range(1, TOTAL):
            net_i: Network = comm.recv(source=i, tag=12)
            for idx, l in enumerate(net_i.layers):
                if hasattr(net.layers[idx], "W"):
                    net.layers[idx].W += l.W / TOTAL
                if hasattr(net.layers[idx], "dydb"):
                    net.layers[idx].dydb += l.dydb / TOTAL
                if hasattr(net.layers[idx], "dydw"):
                    net.layers[idx].dydw += l.dydw / TOTAL

        optimizer.step(net)
        epoch += 1
        if epoch % 10 == 0:
            print(loss_fn(net(X), y))
        for i in range(1, TOTAL):
            comm.send(net, dest=i, tag=33)

elif rank != 0:
    X_on_worker = comm.recv(source=0, tag=22)
    y_on_worker = comm.recv(source=0, tag=23).reshape((-1, 1))
    assert (
        X_on_worker.shape[0] == y_on_worker.shape[0]
    ), f"Number of samples does not match between X and y on {rank}"

    # worker computes the "output", loss function => gradient
    while epoch < args.epochs:
        net_on_worker = comm.recv(source=0, tag=33)
        output = net_on_worker(X_on_worker)  # forward
        loss = loss_fn(output, y_on_worker)  # loss function
        # print(loss)
        g = loss_fn.grad_fn(output, y_on_worker)
        net_on_worker.backward(g, X_on_worker)
        comm.send(net_on_worker, dest=0, tag=12)
        epoch += 1

# if rank == 0:
#     print(loss_fn(y, net(X)))
