"""Basic functions needed for the module."""
import numpy as np
from scipy.special import softmax as s
from nn.datatype import Tensor


def sigmoid(x: Tensor) -> Tensor:
    """Calculate the sigmoid function of x."""
    return 1 / (1 + np.exp(-x))


def sigmoid_prime(x: Tensor) -> Tensor:
    """Calculate the d/dx of sigmoid function of x."""
    return sigmoid(x) * (1 - sigmoid(x))


def softmax(x: Tensor) -> Tensor:
    """Calculate softmax using scipy."""
    return s(x, axis=1)


def relu(x: np.ndarray):
    return np.maximum(x, 0)


def mse(x: np.ndarray, y: np.ndarray) -> float:
    assert x.shape == y.shape
    return 0.5 * np.mean(x - y) ** 2


def mse_grad(x: np.ndarray, y: np.ndarray) -> np.ndarray:
    assert x.shape == y.shape
    return (x - y) / x.shape[0]
