# TODO: accuracy measure, precision, recall, f1-score

import argparse
import enum
import time
from cgi import test
from collections import defaultdict

import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
from mpi4py import MPI
from sklearn.datasets import make_classification
from sklearn.preprocessing import OneHotEncoder


from proj_code import __worker_coeff, part_coeff
from util import (
    load_data,
    slice_node,
    sigmoid,
    sigmoid_derivative,
    softmax,
    softmax_derivative,
    relu,
    relu_derivative,
)


from nn.layers import Linear, ReLU, Sigmoid
from nn.network import Network

comm = MPI.COMM_WORLD
rank = comm.Get_rank()
size = comm.Get_size()
np.random.seed(42)

parser = argparse.ArgumentParser()

parser.add_argument(
    "--samples", "-s", type=int, help=("Sample size"), default=400
)
parser.add_argument(
    "--features", "-f", type=int, help=("Number of features"), default=10
)
parser.add_argument(
    "--learning-rate",
    "-lr",
    type=float,
    help="Learning rate value. Default: 1e-2",
    default=1e-2,
)

parser.add_argument(
    "--epochs",
    "-e",
    type=int,
    help="Number of epochs",
    default=10,
)

parser.add_argument("--num-nodes", "-nn", default=9, type=int)

parser.add_argument("--num-partitions", "-np", default=2, type=int)

parser.add_argument(
    "--num-output-nodes",
    "-nout",
    type=int,
    help="Number of classes (output nodes)",
    default=4,
)

parser.add_argument("--hidden-nodes", "-hn", default=40, type=int)

parser.add_argument(
    "--recovery-threshold",
    "-rt",
    type=int,
    help="Recovery threshold value",
    default=4,
)
args = parser.parse_args()

hidden_nodes = args.hidden_nodes
outputs = args.num_output_nodes

n_workers = args.num_nodes - 1

count_received = 0
count_sent = 0

worker_task_parameter = args.num_partitions
weight_of_code = worker_task_parameter
num_partitions_per_worker = 2 * weight_of_code

worker_coeff = {
    i + 1: np.array(x)
    for i, x in enumerate(
        __worker_coeff(
            np.log2(n_workers).astype("int32") - 1,
            np.log2(weight_of_code).astype("int32"),
        )
    )
}
data_idx = {
    i + 1: tuple(x)
    for i, x in enumerate(
        part_coeff(
            np.log2(n_workers).astype("int32") - 1,
            np.log2(weight_of_code).astype("int32"),
        )
    )
}
# worker_coeff = {
#     1: (1, 1),
#     2: (1, -1),
#     3: (1, 1),
#     4: (1, -1),
#     5: (1, 1),
#     6: (-1, 1),
#     7: (1, 1),
#     8: (-1, 1),
# }

# "mnist"
input_dir = (
    f"./artificial-data/mnist/"  # {args.samples}x{args.features}/{n_workers}/"
)
rows_per_data_batch = args.samples // (n_workers // weight_of_code)

training_losses = defaultdict(list)
test_losses = defaultdict(list)
test_accuracies = defaultdict(list)
times = []
if rank == 0:

    ohe = OneHotEncoder(
        sparse=False,
    )
    # weights = np.zeros((args.features, outputs))

    # initialize weights
    weights_0 = np.random.randn(args.features, hidden_nodes) / np.sqrt(
        args.features
    )
    weights_1 = np.random.randn(hidden_nodes, outputs) / np.sqrt(hidden_nodes)

    # Worker 1,2 will get D1 D2
    y = load_data(input_dir + "label.dat")
    y = ohe.fit_transform(y.reshape((-1, 1)))
    all_data = []  # (D1, y1), (D2, y2), (D3, y3), (D4, y4)
    for i in range(n_workers // weight_of_code):
        all_data.append(
            (
                np.vstack(
                    [
                        load_data(input_dir + f"{2*i+1}.dat"),
                        load_data(input_dir + f"{2*i+2}.dat"),
                    ]
                ),
                y[i * rows_per_data_batch : (i + 1) * rows_per_data_batch],
            )
        )
    # D1, y1 = (
    #     np.vstack(
    #         [load_data(input_dir + "1.dat"), load_data(input_dir + "2.dat")]
    #     ),
    #     y[:rows_per_data_batch],
    # )
    # D2, y2 = (
    #     np.vstack(
    #         [load_data(input_dir + "3.dat"), load_data(input_dir + "4.dat")]
    #     ),  # X[rows_per_data_batch : 2 * rows_per_data_batch],
    #     y[rows_per_data_batch : 2 * rows_per_data_batch],
    # )
    # D3, y3 = (
    #     np.vstack(
    #         [load_data(input_dir + "5.dat"), load_data(input_dir + "6.dat")]
    #     ),  # X[2 * rows_per_data_batch : 3 * rows_per_data_batch],
    #     y[2 * rows_per_data_batch : 3 * rows_per_data_batch],
    # )
    # D4, y4 = (
    #     np.vstack(
    #         [load_data(input_dir + "7.dat"), load_data(input_dir + "8.dat")]
    #     ),  # X[3 * rows_per_data_batch : 4 * rows_per_data_batch],
    #     y[3 * rows_per_data_batch : 4 * rows_per_data_batch],
    # )
    X_test = load_data(input_dir + "test_data.dat")
    y_test = load_data(input_dir + "label_test.dat")
    y_test = ohe.fit_transform(y_test.reshape((-1, 1)))
    workers = [i + 1 for i in range(n_workers)]
    print("LOADING DATA")
    for _iter in range(0, n_workers, worker_task_parameter):
        for worker in workers[_iter : _iter + worker_task_parameter]:
            X_to_send = []
            y_to_send = []
            for idx, d in enumerate(all_data):
                if data_idx[worker][idx] != 0:
                    X_to_send.append(d[0])
                    y_to_send.append(d[1])
            comm.Send(
                np.vstack(X_to_send),
                dest=worker,
                tag=11,
            )
            comm.Send(
                np.vstack(y_to_send),
                dest=worker,
                tag=21,
            )

    # # Worker 3,4 will get D3 D4
    # for worker in [3, 4]:
    #     comm.Send(
    #         np.vstack([D3, D4]),
    #         dest=worker,
    #         tag=11,
    #     )
    #     comm.Send(
    #         np.vstack([y3, y4]),
    #         dest=worker,
    #         tag=21,
    #     )

    # # Worker 5,6 will get D2 D3
    # for worker in [5, 6]:
    #     comm.Send(
    #         np.vstack([D2, D3]),
    #         dest=worker,
    #         tag=11,
    #     )
    #     comm.Send(
    #         np.vstack([y2, y3]),
    #         dest=worker,
    #         tag=21,
    #     )

    # # Worker 7,8 will get D1 D4
    # for worker in [7, 8]:
    #     comm.Send(
    #         np.vstack([D1, D4]),
    #         dest=worker,
    #         tag=11,
    #     )
    #     comm.Send(
    #         np.vstack([y1, y4]),
    #         dest=worker,
    #         tag=21,
    #     )
else:
    X_on_worker = np.zeros(
        (worker_task_parameter * rows_per_data_batch, args.features)
    )

    y_on_worker = np.zeros(
        (worker_task_parameter * rows_per_data_batch, outputs)
    )

    comm.Recv(X_on_worker, source=0, tag=11)
    comm.Recv(y_on_worker, source=0, tag=21)


if rank == 0:
    for i in range(1, n_workers + 1):  # 9)
        comm.Send(
            # weights,
            weights_0,
            dest=i,
            tag=31,
        )
        comm.Send(
            # weights,
            weights_1,
            dest=i,
            tag=32,
        )
        count_sent += 1

# TODO: add cleanup iteration (for-loop?)
# master iterations: num_epochs * (num_workers / recovery_threshold) +
# num_workers * num_epochs - num_epochs * (num_workers / recovery_threshold)
num_iterations = args.epochs * (n_workers // args.recovery_threshold)

if rank > 0:
    for epoch in range(num_iterations):
        if epoch % min(1, (epoch // 100)) == 0:
            print(f"Rank: {rank} at Epoch: {epoch / min(1, (epoch // 100))}%")
        if epoch != (num_iterations - 1):
            # weights_on_worker = np.zeros((args.features, outputs))
            weights_0_on_worker = np.zeros((args.features, hidden_nodes))
            weights_1_on_worker = np.zeros((hidden_nodes, outputs))

            comm.Recv(weights_0_on_worker, source=0, tag=31)
            comm.Recv(weights_1_on_worker, source=0, tag=32)
        # (N, 10) . (10, 100) -> (N, 100) . (100, 2) -> Loss((N, 2)) -> (N, 1)

        # A_i = X_i . W_i^0, dimensions = (n_rows(X_on_worker), hidden_nodes)
        in_layer_1 = X_on_worker.dot(weights_0_on_worker)

        # Z_i = relu(A_i), dimensions = (n_rows(X_on_worker), hidden_nodes)
        out_layer_1 = relu(in_layer_1)

        # B_i = Z_i . W_i^1, dimensions = (n_rows(X_on_worker), outputs)
        logits = out_layer_1.dot(weights_1_on_worker)

        # dimensions = (n_rows(X_on_worker), outputs)
        y_hat = sigmoid(logits)

        # dimensions = (1,)
        loss_on_worker = np.mean(
            np.log(1 + np.exp(-np.multiply(y_on_worker, y_hat)))
        )
        # np.mean(-y_on_worker * np.log(1 + y_hat))

        # shared by everyone
        # dimesnions = (n_rows(X_on_worker), outputs)
        g_loss_yhat = np.divide(
            -y_on_worker * np.exp(-np.multiply(y_hat, y_on_worker)),
            1 + np.exp(-np.multiply(y_hat, y_on_worker)),
        )
        # -y_on_worker / (1 + y_hat)

        # dimensions = (n_rows(X_on_worker), outputs)
        g_yhat_logits = sigmoid_derivative(logits)

        # w1 computations

        # dimensions = (n_rows(X_on_worker), hidden_nodes)
        g_logits_w1 = out_layer_1

        # dimensions = (hidden_nodes, outputs)
        # we did a trick to sum along datapoints to get correct dimensions
        g_loss_w1 = g_logits_w1.T.dot(g_loss_yhat * g_yhat_logits)

        # w0 computations
        # g_logits_w0 = np.zeros((args.features, hidden_nodes, outputs))
        g_logits_w0_helper = np.zeros(
            (
                args.features,
                hidden_nodes,
            ),
            # dtype=np.float32,
        )
        g_loss_w0_helper = np.zeros(
            (
                args.features,
                hidden_nodes,
            ),
            # dtype=np.float32,
        )
        g_logits_w0_helper = np.einsum(
            "ki,kj,jl->ij",
            X_on_worker,
            relu_derivative(in_layer_1),
            weights_1_on_worker,
        )
        g_loss_w0_helper = np.einsum(
            "ij,kl,kl->ij",
            g_logits_w0_helper,
            g_loss_yhat,
            g_yhat_logits,
        )
        # for i in range(X_on_worker.shape[0]):
        #     for j in range(hidden_nodes):
        #         g_logits_w0_helper[i, :, j] = np.outer(
        #             X_on_worker[i],
        #             (
        #                 leaky_relu_derivative(in_layer_1[i, j])
        #                 * (weights_1_on_worker[j])
        #             ),
        #         )
        #         g_loss_w0_helper[i, :, j] = (
        #             g_logits_w0_helper[i, :, j]
        #             * g_loss_yhat[i]
        #             * g_yhat_logits[i]
        #         )

        g_loss_w0 = g_loss_w0_helper  # np.einsum("kl->kl", )

        training_losses[rank].append(loss_on_worker)

        if epoch == (num_iterations - 1):
            for r, lst in training_losses.items():
                plt.plot(lst)
                plt.savefig(f"figures/{r}_training_loss.png")

        g_slice_0 = slice_node(
            g_loss_w0,
            weight_of_code,
            args.features,
            hidden_nodes,
            slice_where="out",
        )
        g_return_0 = np.zeros((args.features, hidden_nodes // weight_of_code))

        g_slice_1 = slice_node(
            g_loss_w1, weight_of_code, hidden_nodes, outputs, slice_where="in"
        )
        g_return_1 = np.zeros((hidden_nodes // weight_of_code, outputs))

        for i in range(weight_of_code):
            g_return_0 += worker_coeff[rank][i] * g_slice_0[i]
            g_return_1 += worker_coeff[rank][i] * g_slice_1[i]
        # (
        # worker_coeff[rank][0] * g[:, 0 : outputs // 2]
        # + worker_coeff[rank][1] * g[:, outputs // 2 :]
        # )
        # if rank in [1, 3, 5, 7]:
        #     g_return = g[:, 0 : outputs // 2] + g[:, outputs // 2 : outputs]
        # elif rank in [2, 4, 6]:
        #     g_return = g[:, 0 : outputs // 2] - g[:, outputs // 2 : outputs]
        # elif rank == 8:
        #     g_return = -g[:, 0 : outputs // 2] + g[:, outputs // 2 : outputs]
        comm.Send(g_return_0, dest=0, tag=33)
        comm.Send(g_return_1, dest=0, tag=34)

else:
    # if n_workers % args.recovery_threshold:  # complicated
    #     num_iterations = num_iterations - 1
    done_counter = dict({i: 0 for i in range(1, n_workers + 1)})

    orig_start_time = time.time()
    for epoch in range(
        num_iterations * (n_workers // args.recovery_threshold)
    ):
        g0 = np.zeros((args.features, hidden_nodes // weight_of_code))
        g1 = np.zeros((hidden_nodes // weight_of_code, outputs))
        count_sources = 0
        sources = []
        while count_sources < args.recovery_threshold:
            # for worker in range(1, 9):
            st = MPI.Status()
            comm.Probe(status=st)
            worker = st.source
            comm.Recv(
                g0,
                source=worker,
                tag=33,
            )
            comm.Recv(
                g1,
                source=worker,
                tag=34,
            )
            count_received += 1
            sources.append(worker)
            done_counter[worker] += 1
            count_sources += 1
            # weights -= 1e-2 * g
            for i in range(worker_task_parameter):
                # print(weights_0.shape, weights_1.shape, g0.shape, g1.shape)
                weights_0[
                    :,
                    i
                    * (hidden_nodes // worker_task_parameter) : (i + 1)
                    * (hidden_nodes // worker_task_parameter),
                ] -= (
                    args.learning_rate
                    * worker_coeff[worker][i]
                    * g0
                    / (args.samples / num_partitions_per_worker)
                )
                weights_1[
                    i
                    * (hidden_nodes // worker_task_parameter) : (i + 1)
                    * (hidden_nodes // worker_task_parameter),
                    :,
                ] -= (
                    args.learning_rate
                    * worker_coeff[worker][i]
                    * g1
                    / (args.samples / num_partitions_per_worker)
                )
            # weights[:, outputs // 2 : outputs] -= (
            #     args.learning_rate
            #     * worker_coeff[worker][1]
            #     * g
            #     / (args.samples / num_partitions_per_worker)
            # )
        # print(
        #     f"Epoch: {epoch}\tCount of sent:{count_sent}\tCount received: {count_received}\tSources: {sources}"
        # )

        in_layer_1 = relu(X_test.dot(weights_0))

        y_hat_test = sigmoid(in_layer_1.dot(weights_1))

        test_losses[rank].append(
            # np.mean(-y_test * np.log(1 + y_hat_test))
            np.mean(np.log(1 + np.exp(-np.multiply(y_test, y_hat_test))))
        )
        times.append(time.time() - orig_start_time)
        acc = np.argmax(y_hat_test, axis=1) == np.argmax(y_test, axis=1)

        test_accuracies[rank].append(acc.mean())

        if epoch == (num_iterations - 2):
            output_csv = pd.DataFrame(columns=["testing_loss", "times"])

            output_csv["testing_loss"] = test_losses[0]
            output_csv["testing_accuracy"] = test_accuracies[0]
            output_csv["times"] = np.array(times) - times[0]

            output_csv.to_csv("figures/proj_grad_code.csv", index=False)
            for r, lst in test_losses.items():
                plt.plot(lst)
                plt.savefig(f"figures/{r}_testing_loss_projective.png")
                plt.close()
            for r, lst in test_accuracies.items():
                plt.plot(lst)
                plt.savefig(f"figures/{r}_testing_accuracy_projective.png")
                plt.close()
        # if epoch != num_iterations - 1:
        if (
            epoch
            == (n_workers // args.recovery_threshold - 1) * num_iterations
        ):
            comm.Abort()

        for i in sources:
            comm.Send(
                weights_0,
                dest=i,
                tag=31,
            )
            comm.Send(
                weights_1,
                dest=i,
                tag=32,
            )
            count_sent += 1

# if n_workers % args.recovery_threshold != 0:
#     print("here")
#     g = np.zeros((args.features, outputs // 2))
#     count_sources = 0
#     sources = []
#     # args.recovery_threshold below must be different: must be cleanup
#     for worker in range(1, 9):
#         # st = MPI.Status()
#         # comm.Probe(status=st)
#         # worker = st.source
#         comm.Recv(
#             g,
#             source=worker,
#             tag=33,
#         )
#         count_received += 1
#         sources.append(worker)
#         count_sources += 1
#         # weights -= 1e-2 * g
#         weights[:, 0 : outputs // 2] -= 1e-2 * worker_coeff[worker][0] * g
#         weights[:, outputs // 2 : outputs] -= (
#             1e-2 * worker_coeff[worker][1] * g
#         )
#     # print(sources)
#     if epoch != args.epochs - 1:
#         for i in sources:
#             comm.Send(
#                 weights,
#                 dest=i,
#                 tag=31,
#             )
#             count_sent += 1
#     print(f"Count of sent:{count_sent},\tCount received: {count_received}")
