from math import sqrt as sqrt
import numpy as np


def encode_chi(i, j, p):
    return (-1) ** (fold_xor(i & j))


def fold_xor(i):
    x = 0
    y = i
    for j in range(32):
        x = (x + (y % 2)) % 2
        y = y // 2
    return x


def code_mat(l: int, p: int):
    k = 2 ** l
    n = 2 * k
    t = 2 ** p
    code = []
    x = x_mat(p)
    l = l_mat(p)
    u = u_mat(p)
    for i in range(k):
        new_row = []
        for j in range((i // t) * t):
            new_row.append(0)
        for j in range((i // t) * t, ((i // t) + 1) * t):
            new_row.append(x[i % t][j % t])
        for j in range(((i // t) + 1) * t, k):
            new_row.append(0)
        if i < t:
            for j in range(t):
                new_row.append(l[i % t][j % t])
            for j in range(t, k - t):
                new_row.append(0)
            for j in range(k - t, k):
                new_row.append(u[i % t][j % t])
        elif i < k:
            for j in range(((i - t) // t) * t):
                new_row.append(0)
            for j in range(((i - t) // t) * t, (((i - t) // t) + 1) * t):
                new_row.append(u[i % t][j % t])
            for j in range((((i - t) // t) + 1) * t, (((i - t) // t) + 2) * t):
                new_row.append(l[i % t][j % t])
            for j in range((((i - t) // t) + 2) * t, k):
                new_row.append(0)
        code.append(new_row)
    return code


def x_mat(p):
    t = 2 ** p
    code = []
    for i in range(t):
        new_row = []
        for j in range(t):
            new_row.append(encode_chi(i, j, p))
        code.append(new_row)
    return code


def neg_x_mat(p):
    t = 2 ** p
    code = []
    for i in range(t):
        new_row = []
        for j in range(t):
            new_row.append((-1) * encode_chi(i, j, p))
        code.append(new_row)
    return code


def z_mat(p):
    t = 2 ** p
    code = []
    for i in range(t):
        new_row = []
        for j in range(t):
            new_row.append(0)
        code.append(new_row)
    return code


def u_mat(p):
    t = 2 ** p
    left_x = x_mat(p - 1)
    right_x = neg_x_mat(p - 1)
    code = []
    for i in range(t // 2):
        code.append(left_x[i] + right_x[i])
    left_z = z_mat(p - 1)
    right_z = z_mat(p - 1)
    for i in range(t // 2):
        code.append(left_z[i] + right_z[i])
    return code


def l_mat(p):
    t = 2 ** p
    left_z = z_mat(p - 1)
    right_z = z_mat(p - 1)
    code = []
    for i in range(t // 2):
        code.append(left_z[i] + right_z[i])
    left_x = x_mat(p - 1)
    right_x = x_mat(p - 1)
    for i in range(t // 2):
        code.append(left_x[i] + right_x[i])
    return code


def __worker_coeff(l: int, p: int) -> list:
    c_mat = np.transpose(np.array(code_mat(l, p)))
    worker_c = []
    k = len(c_mat) // 2
    for i in range(k):
        worker_c.append([c for c in c_mat[i] if c != 0])
    for i in range(k, 2 * k):
        worker_c.append(
            worker_c[i - k][len(worker_c[i - k]) // 2 : len(worker_c[i - k])]
            + worker_c[i - k][: len(worker_c[i - k]) // 2]
        )
    return worker_c


def part_coeff(l, p):
    c_mat = np.transpose(np.array(code_mat(l, p)))
    part_c = []
    for row in c_mat:
        part_c.append([abs(c) for c in row])
    return part_c


def print_code_mat_then_worker_coeff(l, p):
    print(np.transpose(np.array(code_mat(l, p))))
    w_c = __worker_coeff(l, p)
    for i in w_c:
        print(i)
